﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace TestLibrary
{
    /// <summary>
    /// 스트림 재생기 컨트롤
    /// </summary>
    public partial class StreamPlayerControl: UserControl
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 스트림 시작시 이벤트 - StreamStarted

        /// <summary>
        /// 스트림 시작시 이벤트
        /// </summary>
        public event EventHandler StreamStarted;

        #endregion
        #region 스트림 중단시 이벤트 - StreamStopped

        /// <summary>
        /// 스트림 중단시 이벤트
        /// </summary>
        public event EventHandler StreamStopped;

        #endregion
        #region 스트림 실패시 이벤트 - StreamFailed

        /// <summary>
        /// 스트림 실패시 이벤트
        /// </summary>
        public event EventHandler<StreamFailedEventArgs> StreamFailed;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Delegate
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 콜백 대리자 - CallbackDelegate()

        /// <summary>
        /// 콜백 대리자
        /// </summary>
        [UnmanagedFunctionPointer(CallingConvention.StdCall, SetLastError = true)]
        private delegate void CallbackDelegate();

        #endregion
        #region 에러 콜백 대리자 - ErrorCallbackDelegate(error)

        /// <summary>
        /// 에러 콜백 대리자
        /// </summary>
        /// <param name="error">에러</param>
        [UnmanagedFunctionPointer(CallingConvention.StdCall, SetLastError = true)]
        private delegate void ErrorCallbackDelegate(string error);

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 스트림 재생기 헬퍼
        /// </summary>
        private StreamPlayerHelper helper;

        /// <summary>
        /// 스트림 시작시 콜백 대리자
        /// </summary>
        private Delegate streamStartedCallbackDelegate;

        /// <summary>
        /// 스트림 중단시 콜백 대리자
        /// </summary>
        private Delegate streamStoppedCallbackDelegate;

        /// <summary>
        /// 스트림 실패시 콜백 대리자
        /// </summary>
        private Delegate streamFailedCallbackDelegate;

        /// <summary>
        /// 리소스 해제 여부
        /// </summary>
        private bool disposed;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 스트램 재생시 헬퍼 - Helper

        /// <summary>
        /// 스트램 재생시 헬퍼
        /// </summary>
        private StreamPlayerHelper Helper
        {
            get
            {
                return this.helper;
            }
        }

        #endregion
        #region 재생 여부 - IsPlaying

        /// <summary>
        /// 재생 여부
        /// </summary>
        [Browsable(false)]
        public Boolean IsPlaying { get; private set; }

        #endregion
        #region 비디오 크기 - VideoSize

        /// <summary>
        /// 비디오 크기
        /// </summary>
        [Browsable(false)]
        public Size VideoSize
        {
            get
            {
                return IsPlaying ? this.helper.GetFrameSize() : new Size(0, 0);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - StreamPlayerControl()

        /// <summary>
        /// 생성자
        /// </summary>
        public StreamPlayerControl()
        {
            InitializeComponent();

            this.helper = GetStreamPlayerHelper();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 재생 시작하기 - StartPlay(uri, connectionTimeout, transport, flag)

        /// <summary>
        /// 재생 시작하기
        /// </summary>
        /// <param name="uri">URI</param>
        /// <param name="connectionTimeout">연결 타임아웃</param>
        /// <param name="transport">RTSP 전송</param>
        /// <param name="flag">RTSP 플래그</param>
        public void StartPlay(Uri uri, TimeSpan connectionTimeout, RTSPTransport transport, RTSPFlag flag)
        {
            if(IsPlaying)
            {
                Stop();
            }

            this.helper.StartPlay(uri.IsFile ? uri.LocalPath : uri.ToString(), connectionTimeout, transport, flag);
        }

        #endregion
        #region 재생 시작하기 - StartPlay(uri)

        /// <summary>
        /// 재생 시작하기
        /// </summary>
        /// <param name="uri">URI</param>
        public void StartPlay(Uri uri)
        {
            StartPlay(uri, TimeSpan.FromSeconds(5.0), RTSPTransport.Undefined, RTSPFlag.None);
        }

        #endregion
        #region 현재 프레임 구하기 - GetCurrentFrame()

        /// <summary>
        /// 현재 프레임 구하기
        /// </summary>
        /// <returns>비트맵</returns>
        public Bitmap GetCurrentFrame()
        {
            if(!IsPlaying)
            {
                throw new InvalidOperationException();
            }

            return this.helper.GetCurrentFrame();
        }

        #endregion
        #region 중단하기 - Stop()

        /// <summary>
        /// 중단하기
        /// </summary>
        public void Stop()
        {
            if(!IsPlaying)
            {
                throw new InvalidOperationException();
            }

            this.helper.Stop();

            IsPlaying = false;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 리소스 해제하기 - Dispose(disposing)

        /// <summary>
        /// 리소스 해제하기
        /// </summary>
        /// <param name="disposing">해제 여부</param>
        protected override void Dispose(bool disposing)
        {
            if(this.disposed)
            {
                return;
            }

            if(disposing && (this.helper != null))
            {
                if(IsPlaying)
                {
                    Stop();
                }

                this.helper.Uninitialize();

                this.helper.Dispose();
            }

            if(disposing && (this.components != null))
            {
                this.components.Dispose();
            }

            this.disposed = true;

            base.Dispose(disposing);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 스트림 시작시 이벤트 발생시키기 - FireStreamStartedEvent()

        /// <summary>
        /// 스트림 시작시 이벤트 발생시키기
        /// </summary>
        private void FireStreamStartedEvent()
        {
            IsPlaying = true;

            StreamStarted?.Invoke(this, EventArgs.Empty);
        }

        #endregion
        #region 스트림 중단시 이벤트 발생시키기 - FireStreamStoppedEvent()

        /// <summary>
        /// 스트림 중단시 이벤트 발생시키기
        /// </summary>
        private void FireStreamStoppedEvent()
        {
            IsPlaying = false;

            StreamStopped?.Invoke(this, EventArgs.Empty);
        }

        #endregion
        #region 스트림 실패시 이벤트 발생시키기 - FireStreamFailedEvent(error)

        /// <summary>
        /// 스트림 실패시 이벤트 발생시키기
        /// </summary>
        /// <param name="error">에러</param>
        private void FireStreamFailedEvent(string error)
        {
            IsPlaying = false;

            StreamFailed?.Invoke(this, new StreamFailedEventArgs(error));
        }

        #endregion

        #region 스트림 재생기 헬퍼 구하기 - GetStreamPlayerHelper()

        /// <summary>
        /// 스트림 재생기 헬퍼 구하기
        /// </summary>
        /// <returns>스트림 재생기 헬퍼</returns>
        private StreamPlayerHelper GetStreamPlayerHelper()
        {
            StreamPlayerHelper helper = new StreamPlayerHelper();

            this.streamStartedCallbackDelegate = new CallbackDelegate(FireStreamStartedEvent);
            this.streamStoppedCallbackDelegate = new CallbackDelegate(FireStreamStoppedEvent);
            this.streamFailedCallbackDelegate  = new ErrorCallbackDelegate(FireStreamFailedEvent);

            StreamPlayerParameter parameter = new StreamPlayerParameter
            {
                WindowHandle                = Handle,
                StreamStartedCallbackHandle = Marshal.GetFunctionPointerForDelegate(streamStartedCallbackDelegate),
                StreamStoppedCallbackHandle = Marshal.GetFunctionPointerForDelegate(streamStoppedCallbackDelegate),
                StreamFailedCallbackHandle  = Marshal.GetFunctionPointerForDelegate(streamFailedCallbackDelegate)
            };

            helper.Initialize(parameter);

            return helper;
        }

        #endregion
    }
}